@@ -19,26 +19,23 @@
 #include <cupt/cache/binarypackage.hpp>
 #include <cupt/cache/binaryversion.hpp>
 
+#include <internal/versionparse.hpp>
+
 namespace cupt {
 namespace cache {
 
-BinaryPackage::BinaryPackage(const shared_ptr< const string >& binaryArchitecture, bool allowReinstall)
-	: Package(binaryArchitecture), __allow_reinstall(allowReinstall)
+BinaryPackage::BinaryPackage(const string* binaryArchitecture)
+	: Package(binaryArchitecture)
 {}
 
-shared_ptr< Version > BinaryPackage::_parse_version(const Version::InitializationParameters& initParams) const
+unique_ptr< Version > BinaryPackage::_parse_version(const internal::VersionParseParameters& initParams) const
 {
-	auto version = BinaryVersion::parseFromFile(initParams);
-	if (__allow_reinstall && version->isInstalled())
-	{
-		version->versionString += "~installed";
-	}
-	return version;
+	return internal::parseBinaryVersion(initParams);
 }
 
-bool BinaryPackage::_is_architecture_appropriate(const shared_ptr< const Version >& version) const
+bool BinaryPackage::_is_architecture_appropriate(const Version* version) const
 {
-	auto binaryVersion = static_pointer_cast< const BinaryVersion >(version);
+	auto binaryVersion = static_cast< const BinaryVersion* >(version);
 	if (binaryVersion->isInstalled())
 	{
 		return true;
@@ -47,29 +44,41 @@ bool BinaryPackage::_is_architecture_appropriate(const shared_ptr< const Version
 	return (architecture == "all" || architecture == *_binary_architecture);
 }
 
-vector< shared_ptr< const BinaryVersion > > BinaryPackage::getVersions() const
+vector< const BinaryVersion* > BinaryPackage::getVersions() const
 {
-	auto source = _get_versions();
-	vector< shared_ptr< const BinaryVersion > > result;
+	const auto& source = _get_versions();
+	vector< const BinaryVersion* > result;
 	FORIT(it, source)
 	{
-		result.push_back(static_pointer_cast< const BinaryVersion >(*it));
+		result.push_back(static_cast< const BinaryVersion* >(it->get()));
 	}
 	return result;
 }
 
-shared_ptr< const BinaryVersion > BinaryPackage::getInstalledVersion() const
+const BinaryVersion* BinaryPackage::getInstalledVersion() const
 {
-	auto source = getVersions();
-	if (!source.empty() && source[0]->isInstalled())
+	const auto& source = _get_versions();
+	if (!source.empty())
 	{
 		// here we rely on the fact that installed version (if exists) adds first to the cache/package
-		return source[0];
-	}
-	else
-	{
-		return shared_ptr< const BinaryVersion >();
+		auto binaryVersion = static_cast< const BinaryVersion* >(source[0].get());
+		if (binaryVersion->isInstalled())
+		{
+			return binaryVersion;
+		}
 	}
+
+	return nullptr;
+}
+
+auto BinaryPackage::begin() const -> iterator
+{
+	return iterator(_get_versions().begin());
+}
+
+auto BinaryPackage::end() const -> iterator
+{
+	return iterator(_get_versions().end());
 }
 
 }
