@@ -35,6 +35,7 @@ class ReleaseLimits;
 
 using std::list;
 using std::unordered_map;
+using std::unique_ptr;
 
 using boost::xpressive::sregex;
 using boost::xpressive::smatch;
@@ -45,32 +46,34 @@ class CacheImpl
  public:
 	struct PrePackageRecord
 	{
-		size_t offset;
+		uint32_t offset;
 		const pair< shared_ptr< const ReleaseInfo >, shared_ptr< File > >* releaseInfoAndFile;
 	};
+	typedef unordered_map< string, vector< PrePackageRecord > > PrePackageMap;
  private:
 	typedef Cache::IndexEntry IndexEntry;
 	typedef Cache::ExtendedInfo ExtendedInfo;
 	struct TranslationPosition
 	{
-		shared_ptr< File > file;
-		size_t offset;
+		File* file;
+		uint32_t offset;
 	};
 
-	map< string, set< const string* > > canProvide;
-	mutable unordered_map< string, shared_ptr< Package > > binaryPackages;
-	mutable unordered_map< string, shared_ptr< Package > > sourcePackages;
-	map< string, TranslationPosition > translations;
-	mutable unordered_map< string, vector< shared_ptr< const BinaryVersion > > > getSatisfyingVersionsCache;
+	unordered_map< string, vector< const string* > > canProvide;
+	mutable unordered_map< string, unique_ptr< Package > > binaryPackages;
+	mutable unordered_map< string, unique_ptr< Package > > sourcePackages;
+	unordered_map< string, TranslationPosition > translations;
+	mutable unordered_map< string, vector< const BinaryVersion* > > getSatisfyingVersionsCache;
 	shared_ptr< PinInfo > pinInfo;
-	mutable map< shared_ptr< const Version >, ssize_t > pinCache;
+	mutable map< const Version*, ssize_t > pinCache;
 	map< string, shared_ptr< ReleaseInfo > > releaseInfoCache;
+	list< RequiredFile > translationFileStorage;
 	smatch* __smatch_ptr;
 
-	shared_ptr< Package > newSourcePackage(const string&) const;
-	shared_ptr< Package > newBinaryPackage(const string&) const;
-	shared_ptr< Package > preparePackage(unordered_map< string, vector< PrePackageRecord > >&,
-			unordered_map< string, shared_ptr< Package > >&, const string&,
+	Package* newSourcePackage() const;
+	Package* newBinaryPackage() const;
+	Package* preparePackage(unordered_map< string, vector< PrePackageRecord > >&,
+			unordered_map< string, unique_ptr< Package > >&, const string&,
 			decltype(&CacheImpl::newBinaryPackage)) const;
 	shared_ptr< ReleaseInfo > getReleaseInfo(const Config&, const IndexEntry&);
 	void parseSourceList(const string& path);
@@ -79,17 +82,18 @@ class CacheImpl
 			shared_ptr< const ReleaseInfo >, const string&);
 	void processTranslationFiles(const IndexEntry&, const string&);
 	void processTranslationFile(const string& path, const string&);
-	vector< shared_ptr< const BinaryVersion > > getSatisfyingVersions(const Relation&) const;
+	vector< const BinaryVersion* > getSatisfyingVersionsNonCached(const Relation&) const;
+	vector< const BinaryVersion* > getSatisfyingVersionsNonCached(const RelationExpression&) const;
+	ssize_t computePin(const Version*, const BinaryPackage*) const;
  public:
 	shared_ptr< const Config > config;
-	shared_ptr< const string > binaryArchitecture;
-	vector< shared_ptr< sregex > > packageNameRegexesToReinstall;
-	shared_ptr< const system::State > systemState;
+	unique_ptr< const string > binaryArchitecture;
+	unique_ptr< const system::State > systemState;
 	vector< IndexEntry > indexEntries;
 	vector< shared_ptr< const ReleaseInfo > > sourceReleaseData;
 	vector< shared_ptr< const ReleaseInfo > > binaryReleaseData;
-	mutable unordered_map< string, vector< PrePackageRecord > > preSourcePackages;
-	mutable unordered_map< string, vector< PrePackageRecord > > preBinaryPackages;
+	mutable PrePackageMap preSourcePackages;
+	mutable PrePackageMap preBinaryPackages;
 	list< pair< shared_ptr< const ReleaseInfo >, shared_ptr< File > > >
 			releaseInfoAndFileStorage;
 	ExtendedInfo extendedInfo;
@@ -100,12 +104,12 @@ class CacheImpl
 	void processIndexEntries(bool, bool);
 	void parsePreferences();
 	void parseExtendedStates();
-	shared_ptr< const BinaryPackage > getBinaryPackage(const string& packageName) const;
-	shared_ptr< const SourcePackage > getSourcePackage(const string& packageName) const;
-	ssize_t getPin(const shared_ptr< const Version >&, const std::function< string () >&) const;
-	pair< string, string > getLocalizedDescriptions(const shared_ptr< const BinaryVersion >&) const;
+	const BinaryPackage* getBinaryPackage(const string& packageName) const;
+	const SourcePackage* getSourcePackage(const string& packageName) const;
+	ssize_t getPin(const Version*, const std::function< const BinaryPackage* () >&) const;
+	string getLocalizedDescription(const BinaryVersion*) const;
 	void processProvides(const string*, const char*, const char*);
-	vector< shared_ptr< const BinaryVersion > > getSatisfyingVersions(const RelationExpression&) const;
+	vector< const BinaryVersion* > getSatisfyingVersions(const RelationExpression&) const;
 };
 
 }
