@@ -37,6 +37,8 @@ using std::list;
 using std::unique_ptr;
 using std::set;
 
+struct BrokenPair;
+
 class NativeResolverImpl
 {
 	typedef Resolver::Reason Reason;
@@ -45,39 +47,40 @@ class NativeResolverImpl
 	typedef Resolver::SynchronizationReason SynchronizationReason;
 	typedef Resolver::RelationExpressionReason RelationExpressionReason;
 	typedef Solution::Action Action;
+	typedef vector< unique_ptr< Action > > ActionContainer;
 
 	shared_ptr< const Config > __config;
 	shared_ptr< const Cache > __cache;
-	set< string > __manually_modified_package_names;
+	map< string, bool > __auto_status_overrides;
 	unique_ptr< SolutionStorage > __solution_storage;
 	ScoreManager __score_manager;
 	AutoRemovalPossibility __auto_removal_possibility;
 
-	map< string, shared_ptr< const BinaryVersion > > __old_packages;
+	map< string, const BinaryVersion* > __old_packages;
 	map< string, dg::InitialPackageEntry > __initial_packages;
-	RelationLine __satisfy_relation_expressions;
-	RelationLine __unsatisfy_relation_expressions;
+
+	vector< dg::UserRelationExpression > p_userRelationExpressions;
 
 	DecisionFailTree __decision_fail_tree;
 	bool __any_solution_was_found;
 
 	void __import_installed_versions();
 	void __import_packages_to_reinstall();
-	bool __prepare_version_no_stick(const shared_ptr< const BinaryVersion >&,
+	bool __prepare_version_no_stick(const BinaryVersion*,
 			dg::InitialPackageEntry&);
-	float __get_version_weight(const shared_ptr< const BinaryVersion >&) const;
-	float __get_action_profit(const shared_ptr< const BinaryVersion >&,
-			const shared_ptr< const BinaryVersion >&) const;
-	AutoRemovalPossibility::Allow __is_candidate_for_auto_removal(const dg::Element*);
+	float __get_version_weight(const BinaryVersion*) const;
+	float __get_action_profit(const BinaryVersion*, const BinaryVersion*) const;
+
+	bool p_computeTargetAutoStatus(const string&, const Solution&, const dg::Element*) const;
+	AutoRemovalPossibility::Allow p_isCandidateForAutoRemoval(const Solution&, const dg::Element*);
 	bool __clean_automatically_installed(Solution&);
-	void __require_strict_relation_expressions();
+
 	void __pre_apply_action(const Solution&, Solution&, unique_ptr< Action > &&, size_t);
 	void __calculate_profits(vector< unique_ptr< Action > >& actions) const;
 	void __pre_apply_actions_to_solution_tree(
 			std::function< void (const shared_ptr< Solution >&) > callback,
 			const shared_ptr< Solution >&, vector< unique_ptr< Action > >&);
 
-	void __post_apply_action(Solution&);
 	void __final_verify_solution(const Solution&);
 
 	bool __makes_sense_to_modify_package(const Solution&, const dg::Element*,
@@ -88,21 +91,19 @@ class NativeResolverImpl
 	void __add_actions_to_fix_dependency(vector< unique_ptr< Action > >&, const Solution&,
 			const dg::Element*);
 	void __prepare_reject_requests(vector< unique_ptr< Action > >& actions) const;
+	void __fillSuggestedPackageReasons(const Solution&, const string&,
+			Resolver::SuggestedPackage&, const dg::Element*, map< const dg::Element*, size_t >&) const;
 	Resolver::UserAnswer::Type __propose_solution(
 			const Solution&, Resolver::CallbackType, bool);
 
-	void __generate_possible_actions(vector< unique_ptr< Action > >*, const Solution&,
-			const dg::Element*, const dg::Element*, bool);
-
-	static const string __dummy_package_name;
+	void __fill_and_process_introduced_by(const Solution&, const BrokenPair&, ActionContainer* actionsPtr);
+	void __generate_possible_actions(vector< unique_ptr< Action > >*, const Solution&, const BrokenPair&, bool);
  public:
 	NativeResolverImpl(const shared_ptr< const Config >&, const shared_ptr< const Cache >&);
 
-	void installVersion(const shared_ptr< const BinaryVersion >&);
-	void satisfyRelationExpression(const RelationExpression&);
-	void unsatisfyRelationExpression(const RelationExpression&);
-	void removePackage(const string& packageName);
+	void satisfyRelationExpression(const RelationExpression&, bool, const string&, RequestImportance, bool);
 	void upgrade();
+	void setAutomaticallyInstalledFlag(const string& packageName, bool flagValue);
 
 	bool resolve(Resolver::CallbackType);
 };
