@@ -40,7 +40,7 @@ class PinInfo
 					ReleaseVendor, ReleaseVersion, ReleaseComponent, ReleaseLabel, HostName };
 
 			Type type;
-			shared_ptr< sregex > value;
+			sregex value;
 		};
 
 		vector< Condition > conditions;
@@ -48,17 +48,17 @@ class PinInfo
 	};
 
 	shared_ptr< const Config > config;
-	shared_ptr< const system::State > systemState;
+	const system::State* systemState;
 	vector< PinEntry > settings;
 
 	void init();
 	void loadData(const string& path);
-	ssize_t getOriginalAptPin(const shared_ptr< const Version >&) const;
-	void adjustUsingPinSettings(const shared_ptr< const Version >&, ssize_t& priority) const;
+	ssize_t getOriginalAptPin(const Version*) const;
+	void adjustUsingPinSettings(const Version*, ssize_t& priority) const;
  public:
-	PinInfo(const shared_ptr< const Config >&, const shared_ptr< const system::State >&);
+	PinInfo(const shared_ptr< const Config >&, const system::State*);
 
-	ssize_t getPin(const shared_ptr< const Version >&, const string& installedVersionString) const;
+	ssize_t getPin(const Version*, const string& installedVersionString) const;
 };
 
 }
