@@ -21,7 +21,6 @@
 #include <iostream>
 using std::cout;
 using std::endl;
-using std::cerr;
 
 #include <cupt/common.hpp>
 #include <cupt/config.hpp>
@@ -36,12 +35,9 @@ int main()
 	{
 		cupt::shared_ptr< cupt::Config > config(new cupt::Config());
 		cupt::Cache cache(config, false, true, true); // include binary installed packages
-		auto systemState = cache.getSystemState();
-		auto packageNames = systemState->getInstalledPackageNames();
 
-		for (size_t i = 0; i < packageNames.size(); ++i)
+		for (const auto& packageName: cache.getSystemState()->getInstalledPackageNames())
 		{
-			const cupt::string& packageName = packageNames[i];
 			if (cache.isAutomaticallyInstalled(packageName))
 			{
 				continue;
@@ -49,23 +45,23 @@ int main()
 			auto package = cache.getBinaryPackage(packageName);
 			if (!package)
 			{
-				cupt::fatal("no binary package '%s'", packageName.c_str());
+				cupt::fatal2("no binary package '%s'", packageName);
 			}
 			auto installedVersion = package->getInstalledVersion();
 			if (!installedVersion)
 			{
-				cupt::fatal("no installed version in package '%s'", packageName.c_str());
+				cupt::fatal2("no installed version in package '%s'", packageName);
 			}
-			auto policyVersion = cache.getPolicyVersion(package);
-			if (!policyVersion)
+			auto preferredVersion = cache.getPreferredVersion(package);
+			if (!preferredVersion)
 			{
-				cupt::fatal("unable to get policy version for package '%s'", packageName.c_str());
+				cupt::fatal2("unable to get preferred version for package '%s'", packageName);
 			}
 
-			if (installedVersion->versionString != policyVersion->versionString)
+			if (installedVersion != preferredVersion)
 			{
 				cout << packageName << ": " << installedVersion->versionString
-						<< " -> " << policyVersion->versionString << endl;
+						<< " -> " << preferredVersion->versionString << endl;
 			}
 		}
 	}
